@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * Privilege Separation for dhcpcd
- * Copyright (c) 2006-2021 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2023 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -75,6 +75,7 @@
 
 #ifdef HAVE_CAPSICUM
 #include <sys/capsicum.h>
+#include <sys/procdesc.h>
 #include <capsicum_helpers.h>
 #endif
 #ifdef HAVE_UTIL_H
@@ -160,7 +161,7 @@ ps_dropprivs(struct dhcpcd_ctx *ctx)
 #if defined(__NetBSD__) || defined(__DragonFly__) || \
     defined(HAVE_KQUEUE) || defined(HAVE_EPOLL)
 	/* The control proxy *does* need to create new fd's via accept(2). */
-	if (ctx->ps_control_pid != getpid()) {
+	if (ctx->ps_ctl == NULL || ctx->ps_ctl->psp_pid != getpid()) {
 		if (setrlimit(RLIMIT_NOFILE, &rzero) == -1)
 			logerr("setrlimit RLIMIT_NOFILE");
 	}
@@ -318,17 +319,41 @@ ps_rights_limit_stdio(struct dhcpcd_ctx *ctx)
 }
 #endif
 
+#ifdef HAVE_CAPSICUM
+static void
+ps_processhangup(void *arg, unsigned short events)
+{
+	struct ps_process *psp = arg;
+	struct dhcpcd_ctx *ctx = psp->psp_ctx;
+
+	if (!(events & ELE_HANGUP))
+		logerrx("%s: unexpected event 0x%04x", __func__, events);
+
+	logdebugx("%s%s%s exited from PID %d",
+	    psp->psp_ifname, psp->psp_ifname[0] != '\0' ? ": " : "",
+	    psp->psp_name, psp->psp_pid);
+
+	ps_freeprocess(psp);
+
+	if (!(ctx->options & DHCPCD_EXITING))
+		return;
+	if (!(ps_waitforprocs(ctx)))
+		eloop_exit(ctx->ps_eloop, EXIT_SUCCESS);
+}
+#endif
+
 pid_t
-ps_dostart(struct dhcpcd_ctx *ctx,
-    pid_t *priv_pid, int *priv_fd,
-    void (*recv_msg)(void *), void (*recv_unpriv_msg),
-    void *recv_ctx, int (*callback)(void *), void (*signal_cb)(int, void *),
+ps_startprocess(struct ps_process *psp,
+    void (*recv_msg)(void *, unsigned short),
+    void (*recv_unpriv_msg)(void *, unsigned short),
+    int (*callback)(struct ps_process *), void (*signal_cb)(int, void *),
     unsigned int flags)
 {
+	struct dhcpcd_ctx *ctx = psp->psp_ctx;
 	int fd[2];
 	pid_t pid;
 
-	if (xsocketpair(AF_UNIX, SOCK_DGRAM | SOCK_CXNB, 0, fd) == -1) {
+	if (xsocketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_CXNB, 0, fd) == -1) {
 		logerr("%s: socketpair", __func__);
 		return -1;
 	}
@@ -343,44 +368,62 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 	}
 #endif
 
-	switch (pid = fork()) {
+#ifdef HAVE_CAPSICUM
+	pid = pdfork(&psp->psp_pfd, PD_CLOEXEC);
+#else
+	pid = fork();
+#endif
+	switch (pid) {
 	case -1:
+#ifdef HAVE_CAPSICUM
+		logerr("pdfork");
+#else
 		logerr("fork");
+#endif
 		return -1;
 	case 0:
-		*priv_fd = fd[1];
+		psp->psp_pid = getpid();
+		psp->psp_fd = fd[1];
 		close(fd[0]);
 		break;
 	default:
-		*priv_pid = pid;
-		*priv_fd = fd[0];
+		psp->psp_pid = pid;
+		psp->psp_fd = fd[0];
 		close(fd[1]);
 		if (recv_unpriv_msg == NULL)
 			;
-		else if (eloop_event_add(ctx->eloop, *priv_fd,
-		    recv_unpriv_msg, recv_ctx) == -1)
+		else if (eloop_event_add(ctx->eloop, psp->psp_fd, ELE_READ,
+		    recv_unpriv_msg, psp) == -1)
 		{
-			logerr("%s: eloop_event_add", __func__);
+			logerr("%s: eloop_event_add fd %d",
+			    __func__, psp->psp_fd);
 			return -1;
 		}
+#ifdef HAVE_CAPSICUM
+		if (eloop_event_add(ctx->eloop, psp->psp_pfd, ELE_HANGUP,
+		    ps_processhangup, psp) == -1)
+		{
+			logerr("%s: eloop_event_add pfd %d",
+			    __func__, psp->psp_pfd);
+			return -1;
+		}
+#endif
+		psp->psp_started = true;
 		return pid;
 	}
 
 
 #ifdef PLUGIN_DEV
 	/* If we are not the root process, stop listening to devices. */
-	if (priv_fd != &ctx->ps_root_fd)
+	if (ctx->ps_root != psp)
 		dev_stop(ctx);
 #endif
 
 	ctx->options |= DHCPCD_FORKED;
-	if (ctx->fork_fd != -1) {
-		close(ctx->fork_fd);
-		ctx->fork_fd = -1;
-	}
-	pidfile_clean();
-
-	eloop_clear(ctx->eloop);
+	if (ctx->ps_log_fd != -1)
+		logsetfd(ctx->ps_log_fd);
+	eloop_clear(ctx->eloop, -1);
+	eloop_forked(ctx->eloop);
 	eloop_signal_set_cb(ctx->eloop,
 	    dhcpcd_signals, dhcpcd_signals_len, signal_cb, ctx);
 	/* ctx->sigset aready has the initial sigmask set in main() */
@@ -389,16 +432,31 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 		goto errexit;
 	}
 
-	/* We are not root */
-	if (priv_fd != &ctx->ps_root_fd) {
-		ps_freeprocesses(ctx, recv_ctx);
-		if (ctx->ps_root_fd != -1) {
-			close(ctx->ps_root_fd);
-			ctx->ps_root_fd = -1;
-		}
+	if (ctx->fork_fd != -1) {
+		/* Already removed from eloop thanks to above clear. */
+		close(ctx->fork_fd);
+		ctx->fork_fd = -1;
+	}
 
+	/* This process has no need of the blocking inner eloop. */
+	if (!(flags & PSF_ELOOP)) {
+		eloop_free(ctx->ps_eloop);
+		ctx->ps_eloop = NULL;
+	} else
+		eloop_forked(ctx->ps_eloop);
+
+	pidfile_clean();
+	ps_freeprocesses(ctx, psp);
+
+	if (ctx->ps_root != psp) {
+		ctx->options &= ~DHCPCD_PRIVSEPROOT;
+		ctx->ps_root = NULL;
+		if (ctx->ps_log_root_fd != -1) {
+			/* Already removed from eloop thanks to above clear. */
+			close(ctx->ps_log_root_fd);
+			ctx->ps_log_root_fd = -1;
+		}
 #ifdef PRIVSEP_RIGHTS
-		/* We cannot limit the root process in any way. */
 		if (ps_rights_limit_stdio(ctx) == -1) {
 			logerr("ps_rights_limit_stdio");
 			goto errexit;
@@ -406,58 +464,88 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 #endif
 	}
 
-	if (priv_fd != &ctx->ps_inet_fd && ctx->ps_inet_fd != -1) {
-		close(ctx->ps_inet_fd);
-		ctx->ps_inet_fd = -1;
-	}
+	if (ctx->ps_inet != psp)
+		ctx->ps_inet = NULL;
+	if (ctx->ps_ctl != psp)
+		ctx->ps_ctl = NULL;
 
-	if (eloop_event_add(ctx->eloop, *priv_fd, recv_msg, recv_ctx) == -1)
+#if 0
+	char buf[1024];
+	errno = 0;
+	ssize_t xx = recv(psp->psp_fd, buf, sizeof(buf), MSG_PEEK);
+	logerr("pid %d test fd %d recv peek %zd", getpid(), psp->psp_fd, xx);
+#endif
+
+	if (eloop_event_add(ctx->eloop, psp->psp_fd, ELE_READ,
+	    recv_msg, psp) == -1)
 	{
-		logerr("%s: eloop_event_add", __func__);
+		logerr("%d %s: eloop_event_add XX fd %d", getpid(), __func__, psp->psp_fd);
 		goto errexit;
 	}
 
-	if (callback(recv_ctx) == -1)
+	if (callback(psp) == -1)
 		goto errexit;
 
 	if (flags & PSF_DROPPRIVS)
 		ps_dropprivs(ctx);
 
+	psp->psp_started = true;
 	return 0;
 
 errexit:
-	/* Failure to start root or inet processes is fatal. */
-	if (priv_fd == &ctx->ps_root_fd || priv_fd == &ctx->ps_inet_fd)
-		(void)ps_sendcmd(ctx, *priv_fd, PS_STOP, 0, NULL, 0);
-	shutdown(*priv_fd, SHUT_RDWR);
-	*priv_fd = -1;
+	if (psp->psp_fd != -1) {
+		close(psp->psp_fd);
+		psp->psp_fd = -1;
+	}
 	eloop_exit(ctx->eloop, EXIT_FAILURE);
 	return -1;
 }
 
+void
+ps_process_timeout(void *arg)
+{
+	struct dhcpcd_ctx *ctx = arg;
+
+	logerrx("%s: timed out", __func__);
+	eloop_exit(ctx->eloop, EXIT_FAILURE);
+}
+
 int
-ps_dostop(struct dhcpcd_ctx *ctx, pid_t *pid, int *fd)
+ps_stopprocess(struct ps_process *psp)
 {
 	int err = 0;
 
+	if (psp == NULL)
+		return 0;
+
+	psp->psp_started = false;
+
 #ifdef PRIVSEP_DEBUG
-	logdebugx("%s: pid=%d fd=%d", __func__, *pid, *fd);
+	logdebugx("%s: me=%d pid=%d fd=%d %s", __func__,
+	    getpid(), psp->psp_pid, psp->psp_fd, psp->psp_name);
 #endif
 
-	if (*fd != -1) {
-		eloop_event_delete(ctx->eloop, *fd);
-		if (ps_sendcmd(ctx, *fd, PS_STOP, 0, NULL, 0) == -1) {
+	if (psp->psp_fd != -1) {
+		eloop_event_delete(psp->psp_ctx->eloop, psp->psp_fd);
+#if 0
+		if (ps_sendcmd(psp->psp_ctx, psp->psp_fd, PS_STOP, 0,
+		    NULL, 0) == -1)
+		{
+			logerr("%d %d %s %s", getpid(), psp->psp_pid, psp->psp_name, __func__);
+			err = -1;
+		}
+		shutdown(psp->psp_fd, SHUT_WR);
+#else
+		if (shutdown(psp->psp_fd, SHUT_WR) == -1) {
 			logerr(__func__);
 			err = -1;
 		}
-		(void)shutdown(*fd, SHUT_RDWR);
-		close(*fd);
-		*fd = -1;
+#endif
+		psp->psp_fd = -1;
 	}
 
 	/* Don't wait for the process as it may not respond to the shutdown
 	 * request. We'll reap the process on receipt of SIGCHLD. */
-	*pid = 0;
 	return err;
 }
 
@@ -468,6 +556,13 @@ ps_start(struct dhcpcd_ctx *ctx)
 
 	TAILQ_INIT(&ctx->ps_processes);
 
+	/* We need an inner eloop to block with. */
+	if ((ctx->ps_eloop = eloop_new()) == NULL)
+		return -1;
+	eloop_signal_set_cb(ctx->ps_eloop,
+	    dhcpcd_signals, dhcpcd_signals_len,
+	    dhcpcd_signal_cb, ctx);
+
 	switch (pid = ps_root_start(ctx)) {
 	case -1:
 		logerr("ps_root_start");
@@ -558,7 +653,7 @@ ps_managersandbox(struct dhcpcd_ctx *ctx, const char *_pledge)
 	 * If it cannot be opened before chrooting then syslog(3) will fail.
 	 * openlog(3) does not return an error which doubly sucks.
 	 */
-	if (ctx->ps_root_fd == -1) {
+	if (ctx->ps_root == NULL) {
 		unsigned int logopts = loggetopts();
 
 		logopts &= ~LOGERR_LOG;
@@ -607,39 +702,106 @@ ps_stop(struct dhcpcd_ctx *ctx)
 	    ctx->eloop == NULL)
 		return 0;
 
-	r = ps_ctl_stop(ctx);
-	if (r != 0)
-		ret = r;
+	if (ctx->ps_ctl != NULL) {
+		r = ps_ctl_stop(ctx);
+		if (r != 0)
+			ret = r;
+	}
 
-	r = ps_inet_stop(ctx);
-	if (r != 0)
-		ret = r;
+	if (ctx->ps_inet != NULL) {
+		r = ps_inet_stop(ctx);
+		if (r != 0)
+			ret = r;
+	}
 
-	/* We've been chrooted, so we need to tell the
-	 * privileged proxy to remove the pidfile. */
-	ps_root_unlink(ctx, ctx->pidfile);
+	if (ctx->ps_root != NULL) {
+		if (ps_root_stopprocesses(ctx) == -1)
+			ret = -1;
+	}
 
-	r = ps_root_stop(ctx);
-	if (r != 0)
-		ret = r;
-
-	ctx->options &= ~DHCPCD_PRIVSEP;
 	return ret;
 }
 
+bool
+ps_waitforprocs(struct dhcpcd_ctx *ctx)
+{
+	struct ps_process *psp = TAILQ_FIRST(&ctx->ps_processes);
+
+	if (psp == NULL)
+		return false;
+
+	/* Different processes */
+	if (psp != TAILQ_LAST(&ctx->ps_processes, ps_process_head))
+		return true;
+
+	return !psp->psp_started;
+}
+
+int
+ps_stopwait(struct dhcpcd_ctx *ctx)
+{
+	int error = EXIT_SUCCESS;
+
+	if (ctx->ps_eloop == NULL || !ps_waitforprocs(ctx))
+		return 0;
+
+	ctx->options |= DHCPCD_EXITING;
+	if (eloop_timeout_add_sec(ctx->ps_eloop, PS_PROCESS_TIMEOUT,
+	    ps_process_timeout, ctx) == -1)
+		logerr("%s: eloop_timeout_add_sec", __func__);
+	eloop_enter(ctx->ps_eloop);
+
+#ifdef HAVE_CAPSICUM
+	struct ps_process *psp;
+
+	TAILQ_FOREACH(psp, &ctx->ps_processes, next) {
+		if (psp->psp_pfd == -1)
+			continue;
+		if (eloop_event_add(ctx->ps_eloop, psp->psp_pfd,
+		    ELE_HANGUP, ps_processhangup, psp) == -1)
+			logerr("%s: eloop_event_add pfd %d",
+			    __func__, psp->psp_pfd);
+	}
+#endif
+
+	error = eloop_start(ctx->ps_eloop, &ctx->sigset);
+	if (error != EXIT_SUCCESS)
+		logerr("%s: eloop_start", __func__);
+
+	eloop_timeout_delete(ctx->ps_eloop, ps_process_timeout, ctx);
+
+	return error;
+}
+
 void
 ps_freeprocess(struct ps_process *psp)
 {
+	struct dhcpcd_ctx *ctx = psp->psp_ctx;
+
+	TAILQ_REMOVE(&ctx->ps_processes, psp, next);
 
-	TAILQ_REMOVE(&psp->psp_ctx->ps_processes, psp, next);
 	if (psp->psp_fd != -1) {
-		eloop_event_delete(psp->psp_ctx->eloop, psp->psp_fd);
+		eloop_event_delete(ctx->eloop, psp->psp_fd);
 		close(psp->psp_fd);
 	}
 	if (psp->psp_work_fd != -1) {
-		eloop_event_delete(psp->psp_ctx->eloop, psp->psp_work_fd);
+		eloop_event_delete(ctx->eloop, psp->psp_work_fd);
 		close(psp->psp_work_fd);
 	}
+#ifdef HAVE_CAPSICUM
+	if (psp->psp_pfd != -1) {
+		eloop_event_delete(ctx->eloop, psp->psp_pfd);
+		if (ctx->ps_eloop != NULL)
+			eloop_event_delete(ctx->ps_eloop, psp->psp_pfd);
+		close(psp->psp_pfd);
+	}
+#endif
+	if (ctx->ps_root == psp)
+		ctx->ps_root = NULL;
+	if (ctx->ps_inet == psp)
+		ctx->ps_inet = NULL;
+	if (ctx->ps_ctl == psp)
+		ctx->ps_ctl = NULL;
 #ifdef INET
 	if (psp->psp_bpf != NULL)
 		bpf_close(psp->psp_bpf);
@@ -650,12 +812,23 @@ ps_freeprocess(struct ps_process *psp)
 static void
 ps_free(struct dhcpcd_ctx *ctx)
 {
-	struct ps_process *psp;
-	bool stop = ctx->ps_root_pid == getpid();
+	struct ps_process *ppsp, *psp;
+	bool stop;
+
+	if (ctx->ps_root != NULL)
+		ppsp = ctx->ps_root;
+	else if (ctx->ps_ctl != NULL)
+		ppsp = ctx->ps_ctl;
+	else
+		ppsp = NULL;
+	if (ppsp != NULL)
+		stop = ppsp->psp_pid == getpid();
+	else
+		stop = false;
 
 	while ((psp = TAILQ_FIRST(&ctx->ps_processes)) != NULL) {
-		if (stop)
-			ps_dostop(ctx, &psp->psp_pid, &psp->psp_fd);
+		if (stop && psp != ppsp)
+			ps_stopprocess(psp);
 		ps_freeprocess(psp);
 	}
 }
@@ -761,7 +934,6 @@ ps_sendpsmmsg(struct dhcpcd_ctx *ctx, int fd,
 
 	len = writev(fd, iov, iovlen);
 	if (len == -1) {
-		logerr(__func__);
 		if (ctx->options & DHCPCD_FORKED &&
 		    !(ctx->options & DHCPCD_PRIVSEPROOT))
 			eloop_exit(ctx->eloop, EXIT_FAILURE);
@@ -897,7 +1069,8 @@ nobufs:
 }
 
 ssize_t
-ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, uint16_t cmd, int wfd)
+ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, unsigned short events,
+    uint16_t cmd, int wfd)
 {
 	struct sockaddr_storage ss = { .ss_family = AF_UNSPEC };
 	uint8_t controlbuf[sizeof(struct sockaddr_storage)] = { 0 };
@@ -910,32 +1083,33 @@ ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, uint16_t cmd, int wfd)
 		.msg_control = controlbuf, .msg_controllen = sizeof(controlbuf),
 		.msg_iov = iov, .msg_iovlen = 1,
 	};
+	ssize_t len;
 
-	ssize_t len = recvmsg(rfd, &msg, 0);
+	if (!(events & ELE_READ))
+		logerrx("%s: unexpected event 0x%04x", __func__, events);
 
+	len = recvmsg(rfd, &msg, 0);
 	if (len == -1)
 		logerr("%s: recvmsg", __func__);
 	if (len == -1 || len == 0) {
-		if (ctx->options & DHCPCD_FORKED &&
-		    !(ctx->options & DHCPCD_PRIVSEPROOT))
+		if (ctx->options & DHCPCD_FORKED)
 			eloop_exit(ctx->eloop,
-			    len == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+			    len != -1 ? EXIT_SUCCESS : EXIT_FAILURE);
 		return len;
 	}
 
 	iov[0].iov_len = (size_t)len;
 	len = ps_sendcmdmsg(wfd, cmd, &msg);
 	if (len == -1) {
-		logerr("ps_sendcmdmsg");
-		if (ctx->options & DHCPCD_FORKED &&
-		    !(ctx->options & DHCPCD_PRIVSEPROOT))
+		logerr("%s: ps_sendcmdmsg", __func__);
+		if (ctx->options & DHCPCD_FORKED)
 			eloop_exit(ctx->eloop, EXIT_FAILURE);
 	}
 	return len;
 }
 
 ssize_t
-ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
+ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd, unsigned short events,
     ssize_t (*callback)(void *, struct ps_msghdr *, struct msghdr *),
     void *cbctx)
 {
@@ -946,6 +1120,9 @@ ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
 	struct msghdr msg = { .msg_iov = iov, .msg_iovlen = 1 };
 	bool stop = false;
 
+	if (!(events & ELE_READ))
+		logerrx("%s: unexpected event 0x%04x", __func__, events);
+
 	len = read(fd, &psm, sizeof(psm));
 #ifdef PRIVSEP_DEBUG
 	logdebugx("%s: %zd", __func__, len);
@@ -967,6 +1144,7 @@ ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
 	}
 
 	if (stop) {
+		ctx->options |= DHCPCD_EXITING;
 #ifdef PRIVSEP_DEBUG
 		logdebugx("process %d stopping", getpid());
 #endif
@@ -992,6 +1170,8 @@ ps_findprocess(struct dhcpcd_ctx *ctx, struct ps_id *psid)
 	struct ps_process *psp;
 
 	TAILQ_FOREACH(psp, &ctx->ps_processes, next) {
+		if (!(psp->psp_started))
+			continue;
 		if (memcmp(&psp->psp_id, psid, sizeof(psp->psp_id)) == 0)
 			return psp;
 	}
@@ -999,6 +1179,19 @@ ps_findprocess(struct dhcpcd_ctx *ctx, struct ps_id *psid)
 	return NULL;
 }
 
+struct ps_process *
+ps_findprocesspid(struct dhcpcd_ctx *ctx, pid_t pid)
+{
+	struct ps_process *psp;
+
+	TAILQ_FOREACH(psp, &ctx->ps_processes, next) {
+		if (psp->psp_pid == pid)
+			return psp;
+	}
+	errno = ESRCH;
+	return NULL;
+}
+
 struct ps_process *
 ps_newprocess(struct dhcpcd_ctx *ctx, struct ps_id *psid)
 {
@@ -1010,6 +1203,12 @@ ps_newprocess(struct dhcpcd_ctx *ctx, struct ps_id *psid)
 	psp->psp_ctx = ctx;
 	memcpy(&psp->psp_id, psid, sizeof(psp->psp_id));
 	psp->psp_work_fd = -1;
+#ifdef HAVE_CAPSICUM
+	psp->psp_pfd = -1;
+#endif
+
+	if (!(ctx->options & DHCPCD_MANAGER))
+		strlcpy(psp->psp_ifname, ctx->ifv[0], sizeof(psp->psp_name));
 	TAILQ_INSERT_TAIL(&ctx->ps_processes, psp, next);
 	return psp;
 }
